@@ -1,4 +1,4 @@
-/* 
+/*
  * strtod.c --
  *
  *	Source code for the "strtod" library procedure.
@@ -8,12 +8,9 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: strtod.c,v 1.8 2007/04/16 13:36:34 dkf Exp $
  */
 
 #include "tclInt.h"
-#include <ctype.h>
 
 #ifndef TRUE
 #define TRUE 1
@@ -23,12 +20,12 @@
 #define NULL 0
 #endif
 
-static int maxExponent = 511;	/* Largest possible base 10 exponent.  Any
+static const int maxExponent = 511;	/* Largest possible base 10 exponent.  Any
 				 * exponent larger than this will already
 				 * produce underflow or overflow, so there's
 				 * no need to worry about additional digits.
 				 */
-static double powersOf10[] = {	/* Table giving binary powers of 10.  Entry */
+static const double powersOf10[] = {	/* Table giving binary powers of 10.  Entry */
     10.,			/* is 10^2^i.  Used to convert decimal */
     100.,			/* exponents into floating-point numbers. */
     1.0e4,
@@ -63,7 +60,7 @@ static double powersOf10[] = {	/* Table giving binary powers of 10.  Entry */
 
 double
 strtod(
-    CONST char *string,		/* A decimal ASCII floating-point number,
+    const char *string,		/* A decimal ASCII floating-point number,
 				 * optionally preceded by white space. Must
 				 * have form "-I.FE-X", where I is the integer
 				 * part of the mantissa, F is the fractional
@@ -78,8 +75,9 @@ strtod(
 				 * address here. */
 {
     int sign, expSign = FALSE;
-    double fraction, dblExp, *d;
-    register CONST char *p;
+    double fraction, dblExp;
+    const double *d;
+    register const char *p;
     register int c;
     int exp = 0;		/* Exponent read from "EX" field. */
     int fracExp = 0;		/* Exponent that derives from the fractional
@@ -94,7 +92,7 @@ strtod(
     int mantSize;		/* Number of digits in mantissa. */
     int decPt;			/* Number of mantissa digits BEFORE decimal
 				 * point. */
-    CONST char *pExp;		/* Temporarily holds location of exponent in
+    const char *pExp;		/* Temporarily holds location of exponent in
 				 * string. */
 
     /*
@@ -139,7 +137,7 @@ strtod(
      * has more than 18 digits, ignore the extras, since they can't affect the
      * value anyway.
      */
-    
+
     pExp  = p;
     p -= mantSize;
     if (decPt < 0) {
@@ -219,7 +217,7 @@ strtod(
      * by processing the exponent one bit at a time to combine many powers of
      * 2 of 10. Then combine the exponent with the fraction.
      */
-    
+
     if (exp < 0) {
 	expSign = TRUE;
 	exp = -exp;
@@ -231,7 +229,7 @@ strtod(
 	errno = ERANGE;
     }
     dblExp = 1.0;
-    for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
+    for (d = powersOf10; exp != 0; exp >>= 1, ++d) {
 	if (exp & 01) {
 	    dblExp *= *d;
 	}
