@@ -375,6 +375,7 @@ ItclCreateMethod(
         Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
             "bad method name \"", Tcl_GetString(namePtr), "\"",
             (char*)NULL);
+	Tcl_DecrRefCount(namePtr);
         return TCL_ERROR;
     }
 
@@ -479,7 +480,7 @@ ItclCreateMemberFunc(
      *  same name doesn't already exist.
      */
     hPtr = Tcl_CreateHashEntry(&iclsPtr->functions, (char *)namePtr, &newEntry);
-    if (!hPtr) {
+    if (!newEntry) {
         Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
             "\"", Tcl_GetString(namePtr), "\" already defined in class \"",
             Tcl_GetString(iclsPtr->fullNamePtr), "\"",
@@ -1286,8 +1287,6 @@ Itcl_EvalMemberCode(
      */
     if (((mcode->flags & ITCL_IMPLEMENT_OBJCMD) != 0) ||
             ((mcode->flags & ITCL_IMPLEMENT_ARGCMD) != 0)) {
-	Tcl_Namespace *saveNsPtr = Tcl_GetCurrentNamespace(interp);
-	Itcl_SetCallFrameNamespace(interp, imPtr->iclsPtr->nsPtr);
 
         if ((mcode->flags & ITCL_IMPLEMENT_OBJCMD) != 0) {
             result = (*mcode->cfunc.objCmd)(mcode->clientData,
@@ -1306,7 +1305,6 @@ Itcl_EvalMemberCode(
                 ckfree((char*)argv);
 	    }
         }
-	Itcl_SetCallFrameNamespace(interp, saveNsPtr);
     } else {
         if ((mcode->flags & ITCL_IMPLEMENT_TCL) != 0) {
             callbackPtr = Itcl_GetCurrentCallbackPtr(interp);
@@ -1413,68 +1411,116 @@ EquivArgLists(
  *  is not active.
  * ------------------------------------------------------------------------
  */
+
+void
+Itcl_SetContext(
+    Tcl_Interp *interp,
+    ItclObject *ioPtr)
+{
+    int new;
+    Itcl_Stack *stackPtr;
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+            ITCL_INTERP_DATA, NULL);
+    Tcl_HashEntry *hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr, &new);
+    ItclCallContext *contextPtr
+	    = (ItclCallContext *) ckalloc(sizeof(ItclCallContext));
+
+    memset(contextPtr, 0, sizeof(ItclCallContext));
+    contextPtr->ioPtr = ioPtr;
+    contextPtr->refCount = 1;
+
+    if (!new) {
+	Tcl_Panic("frame already has context?!");
+    }
+    
+    stackPtr = (Itcl_Stack *) ckalloc(sizeof(Itcl_Stack));
+    Itcl_InitStack(stackPtr);
+    Tcl_SetHashValue(hPtr, stackPtr);
+
+    Itcl_PushStack(contextPtr, stackPtr);
+}
+
+void
+Itcl_UnsetContext(
+    Tcl_Interp *interp)
+{
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+            ITCL_INTERP_DATA, NULL);
+    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr);
+    Itcl_Stack *stackPtr = (Itcl_Stack *) Tcl_GetHashValue(hPtr);
+    ItclCallContext *contextPtr = Itcl_PopStack(stackPtr);
+
+    if (Itcl_GetStackSize(stackPtr) > 0) {
+	Tcl_Panic("frame context stack not empty!");
+    }
+    Itcl_DeleteStack(stackPtr);
+    ckfree((char *) stackPtr);
+    Tcl_DeleteHashEntry(hPtr);
+    contextPtr->refCount--;
+    if (contextPtr->refCount) {
+	Tcl_Panic("frame context ref count not zero!");
+    }
+    ckfree((char *)contextPtr);
+}
+
 int
 Itcl_GetContext(
     Tcl_Interp *interp,           /* current interpreter */
     ItclClass **iclsPtrPtr,       /* returns:  class definition or NULL */
     ItclObject **ioPtrPtr)        /* returns:  object data or NULL */
 {
-    Tcl_Namespace *activeNs = Tcl_GetCurrentNamespace(interp);
-    Tcl_HashEntry *hPtr;
-    ItclCallContext *callContextPtr;
-    ItclObjectInfo *infoPtr;
+    Tcl_Namespace *nsPtr;
 
-    /*
-     *  Return NULL for anything that cannot be found.
-     */
-    *ioPtrPtr = NULL;
+    /* Fetch the current call frame.  That determines context. */
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
 
-    if (!Itcl_IsClassNamespace(activeNs)) {
-        /*
-         *  If there is no class/object context, return an error message.
-         */
-        Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-            "namespace \"", activeNs->fullName, "\" is not a class namespace",
-            (char*)NULL);
-
-        return TCL_ERROR;
-    }
-    /*
-     *  If the active namespace is a class namespace, then return
-     *  all known info.  See if the current call frame is a known
-     *  object context, and if so, return that context.
-     */
-    infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+    /* Try to map it to a context stack. */
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
             ITCL_INTERP_DATA, NULL);
-    callContextPtr = Itcl_PeekStack(&infoPtr->contextStack);
-    if ((callContextPtr != NULL) && (callContextPtr->imPtr != NULL)) {
-        *iclsPtrPtr = callContextPtr->imPtr->iclsPtr;
-    } else {
-        hPtr = Tcl_FindHashEntry(&infoPtr->namespaceClasses,
-                (char *)activeNs);
-        if (hPtr != NULL) {
-            *iclsPtrPtr = (ItclClass *)Tcl_GetHashValue(hPtr);
-        }
-    }
-    if (*iclsPtrPtr == NULL) {
-        Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-            "namespace \"", activeNs->fullName, "\" is not a class namespace",
-            (char*)NULL);
+    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr);
+    if (hPtr) {
+	/* Frame maps to a context stack. */
+	Itcl_Stack *stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+	ItclCallContext *contextPtr = Itcl_PeekStack(stackPtr);
 
-        return TCL_ERROR;
-    }
+	assert(contextPtr);
 
-    if (callContextPtr == NULL) {
-	/* must be a class namespace without an object */
-	*ioPtrPtr = NULL;
+	if (contextPtr->objectFlags & ITCL_OBJECT_ROOT_METHOD) {
+	    ItclObject *ioPtr = contextPtr->ioPtr;
+
+	    *iclsPtrPtr = ioPtr->iclsPtr;
+	    *ioPtrPtr = ioPtr;
+	    return TCL_OK;
+	}
+
+	*iclsPtrPtr = contextPtr->imPtr ? contextPtr->imPtr->iclsPtr
+		: contextPtr->ioPtr->iclsPtr;
+	*ioPtrPtr = contextPtr->ioPtr ? contextPtr->ioPtr : infoPtr->currIoPtr;
 	return TCL_OK;
     }
-    *ioPtrPtr = callContextPtr->ioPtr;
-    if ((*ioPtrPtr == NULL) && ((*iclsPtrPtr)->nsPtr != NULL)) {
-        /* maybe we are in a constructor try currIoPtr */
-        *ioPtrPtr = (*iclsPtrPtr)->infoPtr->currIoPtr;
+
+    /* Frame has no Itcl context data.  No way to get object context. */
+    *ioPtrPtr = NULL;
+
+    /* Fall back to namespace for possible class context info. */
+    nsPtr = Tcl_GetCurrentNamespace(interp);
+    hPtr = Tcl_FindHashEntry(&infoPtr->namespaceClasses, (char *)nsPtr);
+    if (hPtr) {
+	*iclsPtrPtr = (ItclClass *)Tcl_GetHashValue(hPtr);
+	return TCL_OK;
     }
-    return TCL_OK;
+
+    /* Cannot get any context.  Record an error message. */
+    if (interp) {
+        Tcl_SetObjResult(interp, Tcl_ObjPrintf(
+            "namespace \"%s\" is not a class namespace", nsPtr->fullName));
+    }
+    return TCL_ERROR;
 }
 
 /*
@@ -1808,9 +1854,7 @@ int
 Itcl_ConstructBase(
     Tcl_Interp *interp,       /* interpreter */
     ItclObject *contextObj,   /* object being constructed */
-    ItclClass *contextClass,  /* current class being constructed */
-    int objc,
-    Tcl_Obj *const *objv)
+    ItclClass *contextClass)  /* current class being constructed */
 {
     int result = TCL_OK;
     Tcl_Obj *objPtr;
@@ -1857,7 +1901,7 @@ Itcl_ConstructBase(
 	            contextObj, INT2PTR(0), NULL);
             result = Itcl_NRRunCallbacks(interp, callbackPtr);
 	} else {
-            result = Itcl_ConstructBase(interp, contextObj, iclsPtr, objc, objv);
+            result = Itcl_ConstructBase(interp, contextObj, iclsPtr);
         }
     }
     Tcl_DecrRefCount(objPtr);
@@ -1874,7 +1918,7 @@ ItclConstructGuts(
     ItclClass *contextClass;
 
     /* Ignore syntax error */
-    if (objc != 4) {
+    if (objc != 3) {
 	return TCL_OK;
     }
 
@@ -1883,13 +1927,13 @@ ItclConstructGuts(
 	return TCL_OK;
     }
 
-    contextClass = Itcl_FindClass(interp, Tcl_GetString(objv[3]), 0);
+    contextClass = Itcl_FindClass(interp, Tcl_GetString(objv[2]), 0);
     if (contextClass == NULL) {
 	return TCL_OK;
     }
 
 
-    return Itcl_ConstructBase(interp, contextObj, contextClass, objc, objv);
+    return Itcl_ConstructBase(interp, contextObj, contextClass);
 }
 
 /*
@@ -2138,10 +2182,6 @@ Itcl_CmdAliasProc(
 	if (strcmp(cmdName, "@itcl-builtin-classunknown") == 0) {
 	    return Tcl_FindCommand(interp, "::itcl::builtin::classunknown", NULL, 0);
 	}
-	if (*cmdName == '@') {
-	    return Tcl_FindCommand(interp,
-	            ITCL_NAMESPACE"::methodset::callCCommand", NULL, 0);
-	}
         return NULL;
     }
     clookup = (ItclCmdLookup *)Tcl_GetHashValue(hPtr);
@@ -2268,6 +2308,7 @@ ItclCheckCallMethod(
     Tcl_CallFrame *framePtr,
     int *isFinished)
 {
+    Itcl_Stack *stackPtr;
 
     Tcl_Object oPtr;
     ItclObject *ioPtr;
@@ -2282,13 +2323,12 @@ ItclCheckCallMethod(
     int cObjc;
     int min_allowed_args;
 
+    ItclObjectInfo *infoPtr;
+
     oPtr = NULL;
     hPtr = NULL;
     imPtr = (ItclMemberFunc *)clientData;
     ItclPreserveIMF(imPtr);
-    if (imPtr->codePtr != NULL) {
-        Itcl_PreserveData(imPtr->codePtr);
-    }
     if (imPtr->flags & ITCL_CONSTRUCTOR) {
         ioPtr = imPtr->iclsPtr->infoPtr->currIoPtr;
     } else {
@@ -2390,7 +2430,38 @@ ItclCheckCallMethod(
     if (isNew) {
         Tcl_SetHashValue(hPtr, callContextPtr);
     }
-    Itcl_PushStack(callContextPtr, &imPtr->iclsPtr->infoPtr->contextStack);
+
+    if (framePtr == NULL) {
+	framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    }
+
+    isNew = 0;
+    infoPtr = imPtr->iclsPtr->infoPtr;
+    hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr, &isNew);
+    if (isNew) {
+	stackPtr = (Itcl_Stack *)ckalloc(sizeof(Itcl_Stack));
+	Itcl_InitStack(stackPtr);
+        Tcl_SetHashValue(hPtr, stackPtr);
+    } else {
+	stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    }
+
+    assert (callContextPtr) ;
+    Itcl_PushStack(callContextPtr, stackPtr);
+
+    /* Ugly abuse alert.  Two maps in one table */
+    hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)contextPtr, &isNew);
+    if (isNew) {
+	stackPtr = (Itcl_Stack *)ckalloc(sizeof(Itcl_Stack));
+	Itcl_InitStack(stackPtr);
+        Tcl_SetHashValue(hPtr, stackPtr);
+    } else {
+	stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    }
+
+    Itcl_PushStack(framePtr, stackPtr);
 
     if (ioPtr != NULL) {
         ioPtr->callRefCount++;
@@ -2407,9 +2478,6 @@ ItclCheckCallMethod(
     }
     return result;
 finishReturn:
-    if (imPtr->codePtr != NULL) {
-        Itcl_ReleaseData(imPtr->codePtr);
-    }
     ItclReleaseIMF(imPtr);
     return result;
 }
@@ -2437,10 +2505,31 @@ ItclAfterCallMethod(
     int result;
 
     imPtr = (ItclMemberFunc *)clientData;
-
     callContextPtr = NULL;
     if (contextPtr != NULL) {
-        callContextPtr = Itcl_PopStack(&imPtr->infoPtr->contextStack);
+    ItclObjectInfo *infoPtr = imPtr->infoPtr;
+    Tcl_CallFrame *framePtr;
+    Itcl_Stack *stackPtr;
+	
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)contextPtr);
+    assert(hPtr);
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    framePtr = Itcl_PopStack(stackPtr);
+    if (Itcl_GetStackSize(stackPtr) == 0) {
+	Itcl_DeleteStack(stackPtr);
+	ckfree((char *) stackPtr);
+	Tcl_DeleteHashEntry(hPtr);
+    }
+
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)framePtr);
+    assert(hPtr);
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    callContextPtr = Itcl_PopStack(stackPtr);
+    if (Itcl_GetStackSize(stackPtr) == 0) {
+	Itcl_DeleteStack(stackPtr);
+	ckfree((char *) stackPtr);
+	Tcl_DeleteHashEntry(hPtr);
+    }
     }
     if (callContextPtr == NULL) {
         if ((imPtr->flags & ITCL_COMMON) ||
@@ -2462,6 +2551,7 @@ ItclAfterCallMethod(
      */
     ioPtr = callContextPtr->ioPtr;
     if (ioPtr != NULL) {
+      if (imPtr->iclsPtr) {
         imPtr->iclsPtr->callRefCount--;
         if (imPtr->flags & (ITCL_CONSTRUCTOR | ITCL_DESTRUCTOR)) {
             if ((imPtr->flags & ITCL_DESTRUCTOR) && ioPtr &&
@@ -2475,6 +2565,7 @@ ItclAfterCallMethod(
                     (char *)imPtr->iclsPtr->namePtr, &newEntry);
             }
         }
+      }
         ioPtr->callRefCount--;
         if (ioPtr->flags & ITCL_OBJECT_SHOULD_VARNS_DELETE) {
             ItclDeleteObjectVariablesNamespace(interp, ioPtr);
@@ -2496,9 +2587,6 @@ ItclAfterCallMethod(
     }
     result = call_result;
 finishReturn:
-    if (imPtr->codePtr != NULL) {
-        Itcl_ReleaseData(imPtr->codePtr);
-    }
     ItclReleaseIMF(imPtr);
     return result;
 }
@@ -2516,13 +2604,31 @@ ItclProcErrorProc(
     ItclObject *contextIoPtr;
     ItclClass *currIclsPtr;
     char num[20];
+    Itcl_Stack *stackPtr;
 
+    /* Fetch the current call frame.  That determines context. */
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+
+    /* Try to map it to a context stack. */
     infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
             ITCL_INTERP_DATA, NULL);
-    callContextPtr = Itcl_PeekStack(&infoPtr->contextStack);
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)framePtr);
+    if (hPtr == NULL) {
+	/* Can this happen? */
+	return;
+    }
+
+    /* Frame maps to a context stack. */
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    callContextPtr = Itcl_PeekStack(stackPtr);
+
+    if (callContextPtr == NULL) {
+	return;
+    }
+
     currIclsPtr = NULL;
     objPtr = NULL;
-    if (callContextPtr != NULL) {
+    {
 	imPtr = callContextPtr->imPtr;
         contextIoPtr = callContextPtr->ioPtr;
         objPtr = Tcl_NewStringObj("\n    ", -1);
